program MK454;
{A collection of routines to pre-analyze TCR and Ig deep sequencing data
from 454 instruments to prepare for submission to IMGT. Initially reads in
raw data from large Fasta files.  The data can be parsed into smaller files based on MID numbers,
V family, etc, while stripping out useless sequences, such
 as primer dimers,
very short reads and MID/oligo sequences.

Data are placed in a folder with the appropriate sequence databases.
 Initially the program is for analyzing mouse TCR sequences, but will be
 extended to other sequences later. Initially the databases will contain
 all available sequences, but as sequence data accumulates, the databases
 will be strain specific, starting with C57BL/6J, 129/SvJ and BALB/c. }

Uses CRT,Dos,strings,general,direct;
{CRT, Dos and Strings are Turbo Units with various standard functions
    and Procedures}
{General and Direct are MK Units with additional functions and procedures}
Type
  Char3=Array[1..3]of Char;
  String5=String[5];
  String8=string[8];
  String10=String[10];
  String12=String[12];
  String70=String[70];

  String3=String[3];
  VSeqRec =record
        VFNam:string[10];  {Name of the V Family }
        Fam1:string[12]; {Most 3' sequence that identifies this family}
        Fam1b:integer; {Predicted Distance from begin. of Fam1 to TGT or TGC}
        Fam2:String[12]; {More 5' sequence that identifies this Family}
        Fam2b:Integer; {Predicted Distance from begin. of Fam2 to TGT or TGC}
        Cys1:String[8];  {1st possible Sequence containing the Cys}
        Cys2:String[8];  {2nd possible Sequence containing the Cys}
        pCys:Integer; {Position of CYS in sTGT and sTGC}
        FMemNum:integer; {number of different family members}
        FMemNam:Array[1..11] of String[4]; {member name, e.g 01,07D or 01N}
        FMem1:Array[1..11] of string[11];
        FMem2:Array[1..11] of string[11];
           End;
  JSeqRec = record
        JNam:String[8];{Traj or Trbj Name, e.g. TRAJ0102 or TRBJ0702}
        J1:String[13]; {1st 12base sequence after FGXG unique to this J}
        {It may be useful to move this string before FGXG to avoid PCR Xovers}
        J1n:Integer;
        J2:String[13]; {2nd 12base sequence after FGXG unique to this J}
        J2n:Integer;
        J3:String[9]; {position of FG of J's FGXG}
            End;
VAR
{global variables and data structures]
{Databases}
{This section needs to be tidied up and better annotated}
Dirpath,DirPathF,Subdirpath,INFilename,OutFileName,OutF1Name: String;
Infile,OutFile,TempF:text;
ChrF: file of char;
OutF1, outF2, OutF3, OutF4:text;
Option, contin, T,ans:char;
Comp, OKtoQuit,VDBLoaded,JDBLoaded: Boolean;
ChrA:Array[1..1000] of Char;
CurHead, NextHead, BlankHead:string[80];
CurOrigName, NewOrigName :String[8];
SeqLen,NumofVs,NumOfJs,LenJsearch,LenVsearch,LenCSearch:Integer;
Dataset:Char;
VDBname,JDBName:String[80];
VDBFile,JDBFile:Text;
Vstrain,JStrain:String[8];
VDB:Array [1..40] of VSeqrec;
JDB:Array [1..50] of JSeqrec;
VJAr:Array [1..40] of array [1..50] of LongInt;
C1,C2:String[10];
C1N,C2N:Byte;
{****************************************************************}
Procedure Continue(Var C: Char);
Begin
  Writeln('Press <Enter> to continue or <Esc> to return to Main menu.');
  repeat C:=readkey until C in [#13,#27];
End;
{****************************************************************}

Procedure SplashScreen;
Begin
  Clrscr;
  Gotoxy(20,10);Writeln('MKION VDJ Sequence Analysis Program');
  Writeln;
  Gotoxy(26,12);Writeln('Written by John Kappler');
  Gotoxy(22,13);Writeln('Howard Hughes Medical Institute');
  Gotoxy(25,14);Writeln('Department of Immunology');
  Gotoxy(26,15);Writeln('National Jewish Health');
  Gotoxy(28,16);Writeln('Denver, Colorado');
  Writeln;
  Gotoxy(28,18);Writeln('Version 1.0 (Beta), Feb, 2014');
  Writeln;
  Gotoxy(25,20);Writeln('Press <Enter> to continue');
  Repeat contin:=readkey until contin = #13;
  Clrscr;
end;

{****************************************************************}
Procedure Initialize;
{initiallze the screen and variables}

Begin
  Dirpath:='';
  Infilename:='';
  Outfilename:='';
  Blankhead:='---------------|No Orig.|No V      |No J    |-|0   |0   |';
 {characteres 2-16 are the experiment and sequence desigantion combined
 {            \               \        \          \        \ \    \    \
              1               17       26         37       46 48  53   58
                        15        8          10        8    1  4     4    23
                 seq name       origin     vname    jname  P/N C3B C3E   CDR3 Prot}

  VDBLoaded:=false;
  JdbLoaded:=False;
  SplashScreen;
  Dataset:=#27;
End;

{****************************************************************}
Function FindMatch(B,E:Integer;Search:string;MIS:byte):integer;
{Returns the position of short seq [<15bases] within the current sequence,
  MIS = the number of mismatches allowed}
VAR
J,K,L,LastPos,lenSearch:integer;
Match:Boolean;
MisMatch:Byte;
Begin
  Match:=False;
  J:=B-1;
  LenSearch:=Length(Search);
  LastPos:=E;
  Mismatch:=0; {Current no of mismatches}
  While (NOT Match) and (J<=LastPos) Do {go to next position if there is no match}
  Begin
    {J+1= where the match search starts in CHA.
    J+K = current position in ChrA being examined.
      K = current postion in Search being examined}
    K:=1;
    Match:=true;
    Mismatch:=0;
    {Match will be considered true until the no. of mismatches exceeds MIS}
    {For this postion in CHRA examine the match to Search one base at a time}
    While (MisMatch<=MIS) and (K<=lensearch) do
    Begin
      If (CHRA[j+k]<>search[k]) then
      Begin {A mis match was found}
       Inc(MisMatch);
       If (mismatch>mis) then Match:=False; {stop if too many mismatches}
      ENd;
      Inc(k); {go to next base}
    end;
    inc(j);{go to next position
    If match, stop here and report the postion in Findmatch
    if no MAtch continue on to next postion in CHRA}
  End;
  If Match then Findmatch:= J else FindMatch:=0;
end; {Function FindMatch}
{****************************************************************}
Procedure GetWorkingDirectory;
Begin
  Writeln;
  Writeln('Press <Enter> to pick the working directory');
  repeat contin:=readkey until Contin = #13;

  GetExistingSubDirect(DirPath,True,DirPath);
  DirpathF:=copy(dirpath,1,length(dirpath)-1);
  Chdir(DirpathF);
End;{GetWorkingDirectory}

{****************************************************************}
Procedure LoadVDataBase;
VAR
TStr:String;
I,J,K,Err:Integer;
Begin
  If VDBloaded then exit;
  Writeln;
  Writeln('Press <Enter> for a list of V databases to choose from');
  Repeat contin:=readkey until contin = #13;
  FillChar(VDB,Sizeof(Vdb),#0);
  GetOldFile('*.VDB',VDBName);
  If VDBNAME = '' then
  Begin
    Writeln;Writeln('No V databases present'); Halt;
  End;
  Assign(VDBFile,VDBname);
  Reset(VDBFile);
  I:=1;
  While not EOF(VDBfile) do
  Begin
    {Readin Vdata}
    Readln(VDBFIle,TStr);
    VDB[i].VFNam:=Tstr;

   {seq and distance to Cys of 1st string}
    Readln(VDBFIle,TStr);
    VDB[i].Fam1:=Tstr;

    Readln(VDBFIle,TStr);
    Val(TStr,VDB[i].Fam1b,Err);

   {seq and distance to Cys of 2nd string}
    Readln(VDBFIle,TStr);
    VDB[i].Fam2:=Tstr;

    Readln(VDBFIle,TStr);
    Val(TStr,VDB[i].Fam2b,Err);

    {Position of Cys}
    Readln(VDBFIle,TStr);
    VDB[i].Cys1:=Tstr;

    Readln(VDBFIle,TStr);
    VDB[i].Cys2:=Tstr;

    Readln(VDBFIle,TStr);
    Val(TStr,VDB[i].pCys,Err);

    {Strings to Identify Family Members}
    Readln(VDBFile,Tstr);
    Val(TStr,VDB[i].FMemNum,Err);
    j:= VDB[i].FMemNum;
    For K:= 1 to j do
    Begin
      Readln(VDBFile,tstr);
      VDB[i].FMemNam[k]:=Tstr;
      Readln(VDBFile,Tstr);
      VDB[i].FMem1[k]:=Tstr;
      Readln(VDBFile,Tstr);
      VDB[i].FMem2[k]:=Tstr;
    End;

    If not EoF (VDBfile) then Readln(VDBFile); {Spacer}

    Inc(i);
  End;
   NumOfVs:=i-1;
   Close(VDBFile);
   VDBloaded:=true;
End; {Procedure LoadVDataBase}
{****************************************************************}
Procedure LoadJDataBase;
Var
  I,Err:integer;
  Tstr:String;
Begin
   If JDBloaded then exit;
   Writeln('Press <Enter> for a list of J databases to choose from');
   Repeat contin:=readkey until contin = #13;

   FillChar(JDB, Sizeof(JDB), #0);
   GetOldFile('*.JDB',JDBName);
  If JDBNAME = '' then
  Begin
    Writeln;Writeln('No J databases present'); Halt;
  End;

   Assign(JDBFile,JDBname);
   Reset(JDBFile);
   I:=1;
   While not EOF(JDBfile) do
   Begin
     {Readin Jdata}

     Readln(JDBFIle,TStr);
     JDB[i].JNam:=Tstr;

     Readln(JDBFIle,TStr);
     JDB[i].J1:=Tstr;

     Readln(JDBFIle,TStr);
     Val(TStr,JDB[i].J1n,Err);

     Readln(JDBFIle,TStr);
     JDB[i].J2:=Tstr;

     Readln(JDBFIle,TStr);
     Val(TStr,JDB[i].J2n,Err);

     Readln(JDBFIle,TStr);
     JDB[i].J3:=Tstr;

     If not EOF(JDBFile) then Readln(JDBFile); {Spacer}
     Inc(I);
   End;
   NumOfJs:=i-1;
   Close(JDBfile);
   JDBLoaded:=True;
End; {Procedure LoadDataBases}
{****************************************************************}
Procedure GetDataOption;
{What kind of analysis - mouse/human TCRa or TCRb}
Var
Tstr: String;
TInt: Byte;
I,Err:Integer;
Begin {GetDataOption}
  Clrscr;
  {Get option TCR alpha or beta, Mouse or human}
  Repeat
    Writeln;
    Writeln ('Choose One:');
    Writeln ('1. Mouse V alpha/Jalpha.');
    Writeln ('2. Mouse V beta/Jbeta.');
    Writeln ('3. Human V alpha/Jalpha.');
    Writeln ('4. Human V beta/Jbeta.');
    Writeln ('Esc= back to main menu.');
    Writeln;
    Dataset := Upcase(readkey);
  Until Dataset in ['1','2','3','4',#27];

  {Load the appropriate Databases}
  Case DataSet of
  '1':Begin
            LenVSearch:=180;{Includes the V Search Seqs, CDR3 and J}
            LenJSearch:=37;{Includes all J search seqs and FGXG +3safety bases}
            C1n:=25; {no of Calpha bases to remove from Seq after convertion to + strand}
          End;
  '2','3','4':Begin
            Writeln;
            Writeln('Under development!!');
            Writeln;
            DataSet:=#27;
          End;
  End;
End;{GetDataOption}
{****************************************************************}
Procedure GetStrain;
Var
Opt:char;
Begin
  If Dataset in ['3','4'] then
  begin
    Vstrain := 'HU';
    Jstrain := 'HU';
  end
  else
  Begin
    Writeln;
    Writeln('Choose a mouse strain:');
    Writeln;
    Writeln('1. C57BL/6.');
    Writeln('2. 129.');
    Writeln('3. C57BL/10.');
    Writeln('4. Unknown.');
    Writeln;
    Repeat opt:= Readkey until opt in ['1','2','3','4'];
    Case Opt of
    '1':Begin
          Vstrain:='C57BL/6';
          Jstrain:='C57BL/6';
        end;
    '2':Begin
          Vstrain:='129';
          Jstrain:='129';
        end;
    '3':Begin
          Vstrain:='C57BL/10';
          Jstrain:='C57BL/10';
        end;
    '4':Begin
          Vstrain:='Unknown';
          Jstrain:='Unknown';
        end;
    End;{case}
  End;{if}
end;{getstrain}
{****************************************************************}
Procedure CheckOrigName(OldN:String;Var NewN:string);
VAR
Ans:char;
Blank:Boolean;
I:Byte;
TName:String;
Begin
  NewN:='';
  Writeln;
  Writeln('Options for the Origin field in the output header:');
  Writeln;
  Writeln('1. Leave Blank: -------- ');
  Writeln('2. Enter a specific name, e.g. mouse, human, C57BL/6, etc.');
  If Copy(Oldn,1,5) <> 'rank=' Then {Not Orignal 454 FASTA File}
  Begin
    Writeln ('3. Keep the current name: ',OldN);
    Writeln('Press <1>,<2> or <3>');
    repeat Ans:= ReadKey until Ans in ['1'..'3'];
  End
  Else
  Begin
    Writeln ('Press <1> or <2>');
    repeat Ans:= ReadKey until Ans in ['1','2'];
  End;

  Case Ans of
    '1':NewN:='--------';
    '2':Begin
          Writeln;
          Write('Enter a name for the Origin field (Max 8 characters): ');
          Readln(TName);
          If length(TName)>8 Then
          Begin
            Writeln;
            Writeln('This name is longer than 8 characters!');
            Writeln('Only the first 8 characters will be used.');
            Tname:=Copy(Tname,1,8);
          End
          Else if Length(Tname)<8 then
            For I:=1 to 8-Length(Tname) do Tname:=Tname+' ';
          NewN:=Tname;
        End;
     '3':NewN:=OldN;
  End;
END; {CheckOrigName}

{****************************************************************}
Procedure CompFlip;
{works on global ChrA Array. Seqlen is Global Variable}
VAR
T:Char;
M,N:Integer;
Begin
  {Flip the array}
  N:=Seqlen DIV 2;
  FOR M:=1 TO n DO
  BEGIN
    T:=ChrA[M];
    ChrA[M]:=Chra[Seqlen+1-M];
    ChrA[Seqlen+1-M]:=T;
  END;

  {Change to Complement}
  FOR M:=1 TO Seqlen DO
  BEGIN
    CASE ChrA[M] OF
    'A':CHRA[M]:='T';
    'T':CHRA[M]:='A';
    'C':CHRA[M]:='G';
    'G':CHRA[M]:='C';
    END;
  END;
End;{Comp}

{****************************************************************}
Procedure InputSeq(VAR F:text);
{puts the saved header in the current Header and Fills the Global
sequence char array, ChrA. Sets up the next header if not EOF}

Var
L:word;
c:char;
i,m,n:integer;
Tstr:String;
EOSeq:Boolean;
{CHRA, Curhead, Nexthead AND Seqlen are global variables}
Begin

   {assign the saved header to the current header. This will be the 1st header or
   the header read by input sequence after the previous sequence was input}
   Curhead:=Nexthead;

  {read in this Sequence and Create char array}
  FillChar(ChrA,1000,#0);
  M:=1; {Counter for next open spot in ChrA}
  If NOT Eof(f) then
  BEGIN
    IF OPTION = '1' THEN {in Ion fasta files seq in one line that may be>255 char}
    Begin {load the seq into CHRA, one base at a time}
      while NOT eoln(f) do
      begin
        Read(f,c);
        CHRA[m]:=c;
        inc(m);
      end;
      readln(f,tstr);{go to beginning of next line after end of seq}
      Readln(f,tstr) {read in next header}
    end
    Else
    Begin
      EOseq:=false;
      Readln(f,Tstr);   {read in first line of seq}
      Repeat{Keep reading until you hit the next header or EOF}
        L:=Length(Tstr);                      {Length of Current line}
        Move(tstr[1],CHRA[m],L);              {Add Current line to array}
        M:=M + L;                             {Next available base in ChrA}
        If Not (eof(f)) then                  {exam next line}
        Begin
          Readln(f,Tstr);
          If (Tstr[1]= '>') then  EOseq:=True; {stop -this is the next header}
        End
        else EOSeq:=true;                      {stop -this is EOF}
      Until EOSeq;
    End;                          {Stop when you hit the next Header}
  END;

  {What to do with last TStr?}
  If NOT eof(f) then Nexthead:= Tstr; {Saves the last read line or next header}
  SeqLen:=M-1; {last base is 1<currently available base}

End;{InputSEQ}
{****************************************************************}
Procedure OutputSeq (Start,Finish:integer; VAR f:text);
{Writes any portion of Global Array ChrA to a text Fasta file
CurHead is global}
Var
  I,M,N:integer;
  Tstr:string;
Begin
  Writeln(f,CurHead);
  M:=0; {Line char counter}
  Tstr:='';

  {Write out the seq in chunks of 60}
  For N:= Start to Finish do
  Begin
    TStr:=tSTR + ChrA[N];
    Inc(M);
    If M = 60 then {write out this line. reset M and Tstr}
    Begin
      Writeln(f,Tstr);
      Tstr:='';
      M:=0;
    End;
  End; {tstr contains what ever is left}

  If Tstr<>'' then
  Begin
    Writeln(f,TStr);{Write last partial line of Sequence}
  End;
End;{OutputSeq}

{****************************************************************}
Procedure CleanUpSeq;
VAR
PosContam,Minlen,Maxlen,FirstBase,LastBase,i,j,k,l,m,n:Integer;
IonHeader,TStr,BadSeq:string;
BC:Char;
CROk,BadSeqOK:Boolean;
NumContam,NumClean,Numshort,NumLong,NumOfSeq:Longint;
ScrX,ScrY:Byte;
Begin
  ClrScr;
  Writeln('This routine will clean up your Fasta file for other routines');
  Writeln('Unix style End-of-line (LF only) will be changed to PC style (CR/LF)');
  Writeln('There''s an option to remove contaminating sequences.');
  Writeln('Negative strand sequences will be switched to positive strand.');
  Writeln('Useless short and long sequences can be removed.');
  Writeln;

  Repeat
    Writeln('Press <ENTER> to Continue or <ESC> to Return to the Main Menu');
    contin:=UpCase(readkey);
  Until (contin in [#13,#27]);

  If Contin = #27 then Exit;

  Writeln;
  Writeln('Press <Enter> to load the Main FASTA File');
  repeat contin:=readkey until Contin = #13;
  GetExistingSubDirect(DirPath,True,SubDirPath);

  {Open the main Fasta File}
  GetOldFile(subdirpath+'*.FA',InFilename);
  Assign(Infile, subdirPath+Infilename);
  Reset(infile);

  Writeln;
  Writeln('You can remove contaminating sequences based on the presence');
  Writeln('of a particular sequence.');
  Writeln;
  Writeln('Enter a sequence or press <Enter> to skip this option');
  Writeln;
  Badseq:='';BadseqOK:=True;
  Repeat
    Readln(Badseq);
    If Badseq <> '' then
    Begin
      For I :=1 to length(Badseq) do
      Begin
         Badseq[I] := Upcase(Badseq[I]);
         If Not (Badseq[I] in ['A','C','T','G']) then BadseqOK :=False;
      End;
    End;
  Until BadseqOK;

  {Do Sequences have to be changed to complement}
  Writeln;
  Writeln('MK454 requires positive strand sequences');
  Writeln('Are your sequences positive strand: [Y] or [N]?');
  Writeln;
  Repeat Ans:=upcase(readkey) until (Ans in ['Y','N']);
  If Ans = 'N' then Comp :=true else Comp :=False;
  If Comp then Writeln('The sequences will be converted to the positive strand.');


  {Define Junk sequences}
  Writeln;
  Writeln('What is the length of the SHORTEST useful RAW sequence[0 to 1000]:');
  Repeat Readln(MinLen) until ((minlen>=0) and (Minlen<=1000));

  If MinLen<>0 then
  Begin
    Writeln;
    Writeln('Sequences shorter than ',MinLen,' bases will be removed');
  End
  Else Writeln('No limit on minimum sequence length.');
  Writeln;
  Writeln('What is the length of the LONGEST useful RAW sequence[>0 to 9999]:');
  Repeat Readln(MaxLen) until ((Maxlen>0) and (Maxlen<10000) and (Maxlen>minlen));

  If MaxLen<>9999 then
  Begin
    Writeln;
    Writeln('Sequences longer than ',MaxLen,' bases will be removed');
  End
  Else Writeln('No limit on the maximum sequence length.');
  Writeln;

  Writeln('Press <Enter> to name the output file containing the cleaned up sequences.');
  repeat contin:=readkey until Contin = #13;

  GetAnyFile(SubDirPath+'*.FA',OutF1Name,'FA');
  Assign(outF1,SubDirPath+OutF1Name);
  Rewrite(OutF1);

  Assign(outF2,SubDirPath+'ShortSeq.fa');
  Rewrite(OutF2);

  Assign(outF3,SubDirPath+'LongSeq.fa');
  Rewrite(OutF3);

  If BadSeq <> '' then
  Begin
    Assign(outF4,SubDirPath+'ContmSeq.fa');
    Rewrite(OutF4);
  End;

    {Read in first Header}
  Repeat
    Readln(InFile,TStr);
  Until (Tstr[1] = '>');

  {Correct the origin name in the header, check if this is a 454 or MK454 header}
  CurOrigName:='';
  CheckOrigName(CurOrigName,NewOrigName);

  Writeln;
  Writeln('Processing the sequences from',Infilename);
  Writeln;
  Writeln('This may take some time,  Be Patient!');

  {Determine Postion to write Seq counter}
  Writeln;
  Write('No. of sequences analyzed: ');
  ScrX:=WhereX;ScrY:=whereY;
  NumContam:=0;NumClean:=0;Numshort:=0;NumLong:=0;NumOfSeq:=0;

  CurHead :=Blankhead; {clear the Current header}
  NextHead:=TStr; {Save the first header in Next header}

  Repeat
    InputSeq(infile);
    Firstbase:=1;LastBAse:=Seqlen;
    {CurHead contains 1st Header or Last read Header}

    If Seqlen < MinLen then
    Begin {Write to short file}
      OutputSeq(Firstbase,LastBase,OutF2); {Write to ShortSeq}
      Inc(NumShort);
    End
    Else If Seqlen >Maxlen then
    Begin {Write to long file}
      OutputSeq(Firstbase,LastBase,OutF3); {Write to longSeq}
      Inc(Numlong);
    end
    Else {seq length is okay so this is either Clean or Contaminant}
    Begin
      {Assume no contamination}
      PosContam:=0;

      {If option chosen then determine if contaminant present}
      If BadSeq <> '' then
         PosContam:=FindMatch(FirstBase,LastBase,BadSeq,0);

      {Output to Clean or Contaminant file}
      IF Poscontam <> 0 then {contains the contaminating sequence}
      Begin
        OutputSeq(Firstbase,LastBase,OutF4); {Write to Contseq}
        Inc(NumContam);
      End
      Else {Poscontam=0, so output to Clean Seq}
      Begin
        IF Comp then CompFlip;{Flip ChrA and change to complement}
        {Ion should have removed the Ion oligos and MID sequence
         If Old 454 Va oligos used the 454 B oligo may be present on the 5' end along with the Va oligo.
         This will have to be remove later. }

        {Remove Calpha sequence}
        If Dataset = '1' then lastbase:=lastbase-C1n;

        {Fix the header ion header to fit}
        CurHead:= Copy(CurHead,1,6) + copy(CurHead,9,4) + copy(CurHead,15,4) + ' |' + NewOrigName + Copy(BlankHead,25,33);

        {Output  Header and Sequence}
        OutputSeq(FirstBAse,Lastbase,OutF1);
        Inc(NumClean);
      End;{Clean}
    End;{Contaminated or Clean}

    NumofSeq:=NumClean+Numshort+NumLong+NumContam;
    If (NumOfSeq mod 1000)=0 then
    Begin
      GotoXY(ScrX,SCrY);
      Write(NumofSeq);
    end;
  Until EOF(infile);

  Close(Infile);
  Close(Outf1);
  Close(Outf2);
  Close(outF3);
  If badseq<>'' then Close(OutF4);
  Writeln;
  Writeln;
  Writeln('Finished processing sequences from: ', Infilename);
  Writeln;
  Writeln('Total No. of sequences analyzed = ',NumOfSeq);
  Writeln;
  Writeln(NumClean,' clean sequences are in ', OutF1Name);
  If NumShort>0  then Writeln(NumShort, ' short sequences are in ShortSeq.FA.');
  If NumLong>0  then Writeln(Numlong, ' long sequences are in LongSeq.FA.');
  If NumContam>0 then writeln(NumContam, 'contaminating sequences are in ContmSeq.FA');
  Writeln('It''s a good idea to back up these files at this time');
  Writeln('Press <Enter> to return to Main Menu');

  Repeat contin:=Readkey until (contin = #13);
  Clrscr;
End;
{***************************************************************************}
PROCEDURE Translate(Fb,Lb:integer; Var ps :string);

VAR
  Codon  :Char3;
  AminoAcid,StC,OrfC     :CHAR;
  I,TB,CB,NA  :INTEGER;
  Trans:Boolean;
  Orf:Integer;
{***************************************************************************}
Function TranslateCodon(codon:Char3):Char;
VAR
  AA:Char;

  BEGIN
    CASE Codon[1] OF
      'A':CASE Codon[2] OF
          'A':CASE Codon[3] of
                 'A':AA:='K';
                 'G':AA:='K';
                 'R':AA:='K';
                 'T':AA:='N';
                 'C':AA:='N';
                 'Y':AA:='N';
               ELSE AA:='.';
               END;
           'G':CASE Codon[3] of
                 'A':AA:='R';
                 'G':AA:='R';
                 'R':AA:='R';
                 'T':AA:='S';
                 'C':AA:='S';
                 'Y':AA:='S';
               ELSE AA:='.';
               END;
           'T':CASE Codon[3] of
                 'A':AA:='I';
                 'G':AA:='M';
                 'T':AA:='I';
                 'C':AA:='I';
                 'Y':AA:='I';
               ELSE AA:='.';
               END;
           'C':AA:='T';

          ELSE AA:='.';
          END;
      'G':CASE Codon[2] OF
            'A':CASE Codon[3] OF
                  'A':AA:='E';
                  'G':AA:='E';
                  'R':AA:='E';
                  'T':AA:='D';
                  'C':AA:='D';
                  'Y':AA:='D';
                ELSE AA:='.';
                END;
            'G':AA:='G';
            'T':AA:='V';
            'C':AA:='A';
          ELSE AA:='.';
          END;
      'T':CASE Codon[2] OF
            'A':CASE Codon[3] OF
                  'A':AA:='*';
                  'G':AA:='*';
                  'R':AA:='*';
                  'T':AA:='Y';
                  'C':AA:='Y';
                  'Y':AA:='Y';
                ELSE AA:='.';
                END;
            'G':CASE Codon[3] OF
                  'A':AA:='*';
                  'G':AA:='W';
                  'T':AA:='C';
                  'C':AA:='C';
                  'Y':AA:='C';
                ELSE AA:='.';
                END;
            'T':CASE Codon[3] OF
                  'A':AA:='L';
                  'G':AA:='L';
                  'R':AA:='L';
                  'T':AA:='F';
                  'C':AA:='F';
                  'Y':AA:='F';
                ELSE AA:='.';
                END;
            'C':AA:='S';
          ELSE AA:='.';
          END;
      'C':CASE Codon[2] OF
            'A':CASE Codon[3] OF
                  'A':AA:='Q';
                  'G':AA:='Q';
                  'R':AA:='Q';
                  'T':AA:='H';
                  'C':AA:='H';
                  'Y':AA:='H';
                ELSE AA:='.';
                END;
            'G':AA:='R';
            'T':AA:='L';
            'C':AA:='P';
          ELSE AA:='.';
          END;
      '-':AA:='-';
    ELSE AA:='.';
    END;
  TranslateCodon:=AA;
  END; {TranslateCodon1 ------------------------------------------------------}

BEGIN
  Cb:=Fb;
  ps:='';
  Repeat
    Move(ChrA[cb],codon[1],3);
    AminoAcid:=TranslateCodon(Codon);
    pS:=pS+aminoacid;
    CB:=CB+3;
  Until CB>Lb;
END;
{***************************************************************************}
Procedure IdentifyVJ;
VAR
  NumOfSeq,NumP,NumNonP,NumOOF,NumNoV,NumNoJ,NumNoCys,NumNoFGXG,NumTerm:Longint;
  JN,{J number in the database}
  VN, {V number in tne database}
  JB, {postion of First J germline base in CDR3}
  i,j,k,l,m,n:Integer;
  ScrX,ScrY:Byte;
  FamMemFound,JFound,Vfound,CysFound,FGXGFound,GoodCDR3,Inframe,Term,Productive:Boolean;
  TJ1,TJ2:String[50];
  CDR3DNA:String;
  CDR3Pro:String;
  CysMatch,CDR3B,CDR3E,Jpos,Jpos1,Jpos2,JposN,FPos,Vpos1,Vpos2,Cpos,SB,SE:Integer;
  JSStr1,JSStr2:String[13];
  VSStr1,VSStr2:String[13];
  sCDR3:string[6];
  Tstr:String;
  NamV:String[10];
  NamJ:String[8];
  PNP:Char;
  Scdr3B,Scdr3E:string[4];

Begin
  Clrscr;
  Writeln('This routine will read in your Fasta (.fa) files');
  Writeln('and identify V family and J usage as well as in-frame CDR3.');
  Writeln('The results are stored in the header for each sequence.');
  Writeln('This version does not identify family members. COMING SOON.');
  Writeln('THE CLEAN UP AND MID OPTIONS MUST BE RUN FIRST BECAUSE THE SEQUENCES');
  Writeln('MUST BE + STRAND AND STRIPPED OF MID AND C SEQUENCES FIRST!');

  Repeat
    Writeln('Press <ENTER> to Continue or <ESC> to Return to the Main Menu');
    contin:=UpCase(readkey);
  Until (contin in [#13,#27]);
  If Contin = #27 then Exit;

  LoadVDatabase;
  LoadJDatabase;

    {Open the main Fasta File}
  Infilename := '';
  Writeln;
  Writeln('Press <Enter> to choose a Fasta file to input');
  repeat Contin:=Readkey until contin=#13;
  GetExistingSubDirect(DirPath,True,SubDirPath);
  GetOldFile(subdirPath+'*.FA',InFilename);
  If INfilename = '' then Exit;
  Assign(Infile, subdirPath+Infilename);
  Reset(infile);

  {Open output file}
  Writeln;
  Writeln('Press <Enter> to name the file for output');
  repeat Contin:=Readkey until contin=#13;
  Outfilename:='';
  GetExistingSubDirect(DirPath,True,SubDirPath);

  GetAnyFile(subdirPath+'*.FA',OutFileName,'FA');
  Assign(OutFile,subDirPath+OutFileName);
  Rewrite(OutFile);

   {Find the First header}
  repeat Readln(Infile,Tstr) until (Tstr[1] = '>');
  CurHead:='';
  NextHead:=TStr;

  {Correct the origin name in the header,if required}
  CurOrigName:= Copy(Tstr,17,8);
  CheckOrigName(CurOrigName,NewOrigName);

  Writeln;
  Writeln('Identifying V, J and CDR3. This can take several minutes!');
  Writeln('Be Patient...');
  Writeln;
  Writeln;
  Write('No. of sequences analyzed : ');
  NumOfSeq:=0;
  NumP:=0;NumNonP:=0;NumOOF:=0;NumNoV:=0;
  NumNoJ:=0;NumNoCys:=0;NumNoFGXG:=0;NumTerm:=0;
  ScrX:=wherex;ScrY:=whereY;


  Repeat
    {Read in the sequences and header one at a time}
    InputSeq(Infile);

    {assume all matches false until a match is found}
    JFound:=false;Vfound:=False;Inframe:=False;Productive:=False;
    CysFound:=False;GoodCDR3:=False;Term:=False;CDR3Pro:='';
    CDR3b:=0;CDR3e:=0;
    {Identify the J segment}
    JN:=0;{Counter for current J}
    Jpos1:=0;Jpos2:=0;
    Jfound:=false;
    NamJ:='';
    SB:=SeqLen-(LenJsearch+3);
    SE:=SeqLen;
    Fpos:=0;
    {try the most 5' seq 1st, try to eliminate cross PCRed J's}
    While (Not Jfound) and (JN<NumOfJs) do
    Begin
      Inc(JN);
      Jpos1:=0;
      Jpos1 := findMatch(SB,SE,JDB[JN].J1,1);
      If (Jpos1 <> 0) then {Use most defintive first}
      Begin
        Jfound:=True;
        Fpos:=Jpos1-JDB[JN].J1n;{potential position of last base in CDR3}
      End
    End;

    {If not found use less defintive jseq}
    If (Not Jfound) then JN:=0; {reset JN}
    While (Not Jfound) and (JN<NumOfJs) do
    Begin
      Inc(JN);
      Jpos2:=0;
      Jpos2 := FindMatch(SB,SE,JDB[JN].J2,0);{no mismatches allowed}
      If (Jpos2 <> 0) then
      Begin
        Jfound:=True;
        Fpos:=Jpos2-JDB[JN].J2n;{potential position of last base in CDR3}
      End;
    End;
    {JN now holds the number of the Identified J,
    Fpos the predicted V end of CDR3}
    If Jfound then NamJ:=JDB[JN].JNam else inc(NumNoJ);

    {Identify V segment}
    VN:=0;{Counter for current V}
    Vfound:=false;Vpos1:=0;Vpos2:=0;Cpos:=0;NamV:='';
    SB:= Seqlen-LenVSearch+1;{Just search the 3' end of V}
    If jfound then SE:=Fpos {Stop just before the FGXG}
       Else SE := seqlen;   {Maybe there is no J}

    {try the most 3' seq 1st, try to eliminate cross PCRed V's}
    While (Not Vfound) and (VN<NumOfVs) do
    Begin
      Inc(VN);
      Vpos1 := findMatch(SB,SE,VDB[VN].Fam1,0);
      If Vpos1 <> 0 then
      Begin
        Vfound:=True;
        Cpos:= Vpos1 + VDB[VN].Fam1b;
      End;
    End;

    {if V not found try the more 5' seq}
    If (Not Vfound) then VN:=0; {reset VN}
    While (Not Vfound) and (VN<NumOfVs) do
    Begin
      Inc(VN);
      Vpos2 := FindMatch(SB,SE,VDB[VN].Fam2,0);
      If Vpos2 <> 0 then
      Begin
        Vfound:=True;
        Cpos:= Vpos2 + VDB[VN].Fam2b;
      End;
    End;

    {VN now holds the number of the Current V family}
    {Cpos holds postion of the predicted 1st base of Cys}
    If Vfound then NamV:=VDB[VN].VFNam else Inc(NumNoV);


    If VFound then
    Begin
       {First try to found the Vfam member}

       K:=VDB[VN].FMemNum; I:=0;FamMemFound:=False;

      {try the 1st Fmem seq}
      While (Not FamMemFound) and (I<K) do
      Begin
        Inc(i);
        FamMemFound:=false;
        If VDB[VN].Fmem1[i]<>'NULL' then {This Family has more than one member}
        Begin
          J := findMatch(1,Cpos,VDB[VN].FMem1[i],0);
          If J <> 0 then {Match found. Do 2nd search, if needed}
          Begin
            If VDB[VN].Fmem2[i]<>'NULL' then
            Begin
              j:= findMatch(1,Cpos,VDB[VN].FMem2[i],0);
              If J<>0 then FamMemfound:=True;
            End
            Else FamMemFound:=True;{There is no 2nd search string}
          End
          Else FamMemfound:=false;{The Family member can't be identifed}
        End
        Else FamMemFound:=true;{This Family has only one member}
      End;{While}

      If FamMemFound then
      begin
        NamV:= NamV + VDB[VN].FMemNam[i];
        If Length(NamV) <10 then Repeat Namv:=NamV+' ' until length(NamV) = 10;
      End
      Else NamV:=NamV + '    ';


      {Then find V Cys}
      CysFound:=False;
      {Use Cpos to find a position downstream to Start search for Cys}
      {Now Define a 5base region to look for Cys}
      SB:=Cpos-VDB[VN].pCys+1;
      SE:=SB+4;
      {SB and SE now mark the first and last base in ChrA of the search region}

      {Confirm the presence and position of the Cys in the stretch}
      {1st look in the expected position}
      CysMatch:=FindMatch(sb,se,VDB[VN].cys1,1);
      If (CysMatch <> 0) then CysFound:=True
      else If (VDB[VN].cys2 <> 'NULL') then
      Begin
        CysMatch:=findmatch(sb,se,VDB[VN].cys2,1);
        If (CYSMAtch<>0) then CysFound :=True;
      End;

      {If no match look upstream and down stream 1 base}
      If (NOT Cysfound) then{slide forward one base}
      Begin
        Inc (SE);Inc(SE);
        CysMatch:=FindMatch(sb,se,VDB[VN].cys1,0);
        If (CysMatch <> 0) then CysFound:=True
        else If (VDB[VN].cys2 <> 'NULL') then
        Begin
          CysMatch:=findmatch(sb,se,VDB[VN].cys2,0);
          If (CYSMAtch<>0) then CysFound :=True;
        End;
      end;

      If (Not Cysfound) then{slide back 1 base}
      Begin
        Sb:=Sb-2;SE:=SE-2;
        CysMatch:=FindMatch(sb,se,VDB[VN].cys1,0);
        If (CysMatch <> 0) then CysFound:=True
        else If (VDB[VN].cys2 <> 'NULL') then
        Begin
          CysMatch:=findmatch(sb,se,VDB[VN].cys2,0);
          If (CYSMAtch<>0) then CysFound :=True;
        End;
      End;
      CDR3B:=0;
      IF Cysfound then CDR3B:=Cysmatch+(VDB[VN].pCys-1) else Inc(NumNoCys);
      {CDR3B now holds the position of the first base of the CDR3 loop or 0}
    End;{if Vfound}


    {Now find the F in FGXG}
    FGXGFound:=False;
    If Jfound Then
    Begin
      {Define a region to look for the end of CDR3 in J}
      {thus avoiding the downstream seq errors}

      GoodCDR3:=False;
      {Start from Fpos to find FGXG, if not found Move Fpos back
      or foward 1 base, if still not found than Good CDR3 is false. }
      SB:=Fpos-8; {base  #1 in XXF?}
      SE:=SB+8; {Should define XXF}
      CDR3E:=findmatch(SB,SE,JDB[JN].J3,1); {search with no gaps or insertions}
      If CDR3E=0 then
      Begin
        Dec(SB);Dec(SE);
        CDR3E:=findmatch(SB,SE,JDB[JN].J3,1); {back up one base}
      end;
      If CDR3E=0 then
      Begin
        Inc(SB);Inc(SB);Inc(SE);Inc(SE);
        CDR3E:=findmatch(SB,SE,JDB[JN].J3,1); {go forward one base}
      end;
      If CDR3E<>0 then
      Begin
        FGXGFound:=true;
        CDR3E:=CDR3E+8;
      End else Inc(NumNoFGXG);
    END;{IF JFound}

    {If both ends of the CDR3 are found then GoodCDR3 is true}
    GoodCDR3:=False;
    IF (CysFound and FGXGFound) then GoodCDR3:=True;

    {Now there are 3 possiblities - inframe/product,Inframe/Term or Out of frame}
    If GoodCDR3 then
    Begin
      {Num of bases in repaired J stretch}

       {Save the CDR3 DNA seq in CDR3DNA}
       CDR3DNA:='';
       For I:= CDR3b to CDR3e do CDR3DNA:=CDR3DNA+ChrA[i];

        {Check if in frame}
       InFrame:=False;
       IF (Length(CDR3DNA) mod 3) = 0 then InFrame := true
            else Inc(NumOOF);

       {Translate CDR3 and look for Term codon}
       If InFrame then
       Begin
         Term:=False;CDR3PRO:='';
         Translate (CDR3B,CDR3E,CDR3PRO);
         {If Length(CDR3Pro)<23 then Repeat CDR3Pro:=(CDR3PRO + ' ') until Length(CDR3Pro)=23;}
         If (pos('*',CDR3PRO) <>0) then
         begin
           Term:=True;
           Inc(NumTerm);
         End
         else
         Begin
           Inc (NumP);
           productive:=true;
         End;
       End;{InFrame}
    End;{If GoodCDR3}

    {We now have all of the info to fill out the header}

    {Fill current header}

    {Origin name}
    Move(NewOrigName[1],CurHead[17],8);

    {V family name}
    If Vfound then Move (NamV[1],Curhead[26],length(NamV));

    {J name}
    If Jfound then Move (NamJ[1],CurHead[37],length(NamJ));

    {Productive, OOF, Term or not determined}
    If GoodCDR3 then {in frame-Productive, inframe term or out of frame}
    Begin
      If (inframe and (Not term)) then Curhead[46]:='P' {and CD3Rpro has seq}
      Else if (inframe and term) then Curhead[46]:='T' {and CD3Rpro has seq}
      Else if (not inframe) then
      Begin
        Curhead[46]:='O';
        CDR3Pro:='CDR3 out of frame';
      End;
    End
    Else {not GoodCDR3}
    Begin
      CurHead[46]:='U';{Undetermined}
      If ((not Vfound) or (not JFound)) then
          CDR3Pro:='V +/- J not identifed'
      Else CDR3pro:='Ends of CDR3 not found';
    End;

    Curhead:=Copy(Curhead,1,57)+CDR3Pro;

    {CDR3 DNA Begin and End or 0 |0 if not found}
    Tstr:='    |    ';
    Str(CDR3b,Scdr3b);
    Str(Cdr3e,Scdr3e);
    Move(Scdr3b[1],Tstr[1],Length(scdr3b));
    MOve(Scdr3e[1],Tstr[6],Length(Scdr3e));
    Move(tstr[1],curhead[48],9);

    {Output header and Sequence}
    OutputSeq(1,seqlen,Outfile);

    Inc(NumOfSeq);
    IF NumofSEQ MOD 1000 = 0 then
    begin
      GOtoxy(Scrx,ScrY);
      Write(NumOfSeq:7);
    End;
  Until EOF(infile);{if Not EOF- Go on to next Sequence}

  Writeln; writeln;
  Writeln('Finished the analysis of ',NumofSeq,' sequences.');
  Writeln;
  Writeln('      Productive :',NumP:7);
  Writeln('    Unproductive :',NumofSeq-NumP:7);
  Writeln('      No V found :',NumNoV:7);
  Writeln('      No J found :',NumNoJ:7);
  Writeln(' V Cys not found :',NumNoCys:7);
  Writeln('J FGXG not found :',NumNoFGXG:7);
  Writeln('    Out of Frame :',NumOOF:7);
  Writeln('     Term. Codon :',NumTerm:7);
  Writeln;
  Writeln('Your processed data are in file :',outfilename);
  Writeln('Press <ENTER> to continue');
  Repeat contin:=Readkey until contin=#13;

  Clrscr;
End;{IdentifyVJ}
{****************************************************************}
Procedure MakeExcelVJTable;
VAR
  OutOpt:char;
  I,J,K,L,Err:integer;
  NumOfSeq,NumP,NumN:longint;
  ScrX,ScrY,VN,JN:byte;
  sVN,sJN:String[3];
  tStr:String;
Begin
  Clrscr;
  Writeln('This routine will create an Excel compatible table ');
  Writeln('showing the frequency of VJ combinations.');
  Writeln('You must run the "Identify V,J,CDR3" Option first.');
  Writeln('All sequences with analyzed CDR3s are counted so,');
  Writeln('if you want only productive, etc., parse them out first.');
  Writeln('Sequences with undetermined CDR3s are skipped.');
  Writeln;
  Writeln('Press "Enter" to continue.');
  Repeat contin:=readkey until (contin in [#13,#27]);

  {Get input file}
  Writeln;
    {Open the main Fasta File}
  Infilename := '';
  Writeln;
  Writeln('Press <Enter> to choose a Fasta file to input');
  repeat Contin:=Readkey until contin=#13;
  GetExistingSubDirect(DirPath,True,SubDirPath);

  GetOldFile(subdirpath+'*.FA',InFilename);
  If INfilename = '' then Exit;
  Assign(Infile, subdirPath+Infilename);
  Reset(infile);
  Writeln;

  {If not loaded, load the databases}
  Writeln;
  LoadVDatabase;
  Writeln;
  LoadJDatabase;

  {Get option for output}
  Writeln;
  GetExistingSubDirect(DirPath,True,SubDirPath);
  GetNewFile(Subdirpath+'*.TXT',OutFileName,'TXT');
  Assign(OutFile,subdirpath+OutFilename);
  Rewrite(OutFile);

  {readin headers one at a time and extract info}
  NumofSeq:=0;NumP:=0;NumN:=0;
  fillchar(VJAr,sizeof(vjar),#0);

  Writeln;
  Writeln('Analyzing the sequences.  Be Patient...');
  Write('No. of sequences analyzed: ');
  ScrX:=WhereX; Scry:=WhereY;

  repeat Readln(Infile, Tstr) until (Tstr[1]='>');

  while NOT eof(infile) do
  Begin
    IF Tstr[46]<>'U' then {everything but unanalyzed}
    begin
      sVN:=Tstr[26]+Tstr[27];val(svn,vn,err);
      sJN:=Tstr[37]+Tstr[38];val(sjn,jn,err);
      Inc(vjar[vn][jn]);
      Inc(numofSeq);
      If (NumOfSeq mod 1000) = 0 then
      Begin
        Gotoxy(ScrX,ScrY);
        Write(NumOfSeq:8);
      End;
    end;
    Repeat Readln(Infile,tstr) Until Eof(Infile) or (Tstr[1]='>');
  End;


  {Output the table}
  {Output header}
  Write(outfile,'V/J');
  For I:= 1 to NumofJs do write(outfile,',',Copy(jdb[I].Jnam,3,4));
  Writeln(Outfile);
  For J:= 1 to numofvs do
  begin
    Write(outfile, Copy(Vdb[j].VFnam,3,4));
    For I:=1 to NumOfJs do write(outfile,',',vjar[j][I]);
    Writeln(outfile);
  End;
  {Write summary to screen and log file}
  Writeln;
  Writeln;
  Writeln(NumOfSeq,'Sequences have been analyzed from file',InFileName);
  Writeln('Your data have been output to file: ',OutFilename);
  Continue(contin);
  Close(infile);Close(outfile);
  Clrscr;
End; {MakeExcelVJTable}
{****************************************************************************}
Procedure MakeMKSETFile;
{This is a hidden procedure to convert the first 250 sequences in a Fasta file
  to a file that can be analyzed with MKSET. part of the fasta header will be
  placed in the MKSET comment line}
VAR
  OutOpt:char;
  I,J,K,L,Err:integer;
  NumOfSeq,NumP,NumN:longint;
  ScrX,ScrY,VN,JN:byte;
  sVN,sJN:String[3];
  tStr:String;
  ComStr:String[66];
  SeqNam:String[8];
Begin
  Clrscr;
  Writeln('This routine will create an MKSET File');
  Writeln('with up to the 1st 250 Sequences in a FASTA File');
  Writeln('Press "Enter" to continue.');
  Repeat contin:=readkey until (contin in [#13,#27]);

  {Get input file}
  Writeln;
    {Open the main Fasta File}
  Infilename := '';
  Writeln;
  Writeln('Press <Enter> to choose a Fasta file to input');
  repeat Contin:=Readkey until contin=#13;

  GetOldFile('*.FA',InFilename);
  If INfilename = '' then Exit;
  Assign(Infile, Infilename);
  Reset(infile);
  Writeln;

  {Open Output File}
  Outfilename:=Copy(infilename,1,length(infilename)-3);{strip off .FA}
  Outfilename:=OutFileName+'.SET';
  Assign(OutFile,OutFilename);
  Rewrite(OutFile);

  {readin headers one at a time and extract info}
  NumofSeq:=0;NumP:=0;NumN:=0;

  repeat Readln(Infile, Tstr) until (Tstr[1]='>');{1st Header}

  while NOT eof(infile) and (numofSeq<251) do
  Begin
    Comstr:=Copy(Tstr,17,length(tstr)-16);
    SeqNam:=Copy(Tstr,11,5);
    Writeln(OutFile,SeqNam);
    Writeln(Outfile,'DNA');
    Writeln(OUtfile,ComStr);
    Repeat
      readln(Infile,tstr);
      If (tstr[1]<> '>') then Writeln(outfile,tstr);
    Until EOF(Infile) or (tstr[1] = '>');
    Writeln(OutFile);
    Inc(NumOfSeq);
  end; {while}
  Writeln(OutFile,'END');

  {Write summary to screen and log file}
  Writeln;
  Writeln(NumOfSeq,' Sequences have been analyzed from file ',InFileName);
  Writeln('Your data have been output to file: ',OutFilename);
  Continue(Contin);
  Close(infile);Close(outfile);
  Clrscr;
End; {MakeMKsetFile}
{****************************************************************}
Procedure ExtractSubset;
VAR
  Tstr,S1, S2:string;
  DNAMatchPos:integer;
  ScrX,ScrY,NumS:byte;
  NumOut,TotSeq:longint;
  ORs,ANDs,S1Dna,S2dna,S1Found,S2Found:Boolean;
Begin
  Clrscr;
  Writeln('This routine will extract a subset of sequences from a Fasta file');
  Writeln('based on any text in the sequence headers or the sequences.');
  Writeln('DO THE V, J, CDR3 ANALYSIS FIRST!!');
  Writeln('You can combine two terms with [AND] and [OR] connectors');
  Writeln('For example, entering: |O|[OR]|T| extracts all sequences whose');
  Writeln('CDR3s are either out-of-frame or have a termination codon.');
  Writeln('Entering, AV09[AND]AJ17 extracts all sequences that have both');
  Writeln('TRAV09 and TRAJ17. Enclosed DNA sequences in {} to distinguish');
  Writeln('from header text. See the documentation for more examples and');
  Writeln('the structure of the sequence headers after VJCDR3 analysis.');
  {This procedure should be expanded for more complex searches}

  Writeln;
  Writeln('Press <Enter> to continue or <Esc> to return to Main Menu.');

  Continue(contin);
  If contin = #27 then exit;

  Writeln;
  Writeln('Enter the header text/DNA sequence to search for:');
  Writeln;
  readln(Tstr);
  ORs:=false; ANDs:=False; S1dna:=false; S2dna:=false;
  S1:='';S2:='';Nums:=0;Numout:=0;
  If (pos('[AND]',tstr)=0) and (pos('[OR]',tstr)=0) then
  Begin
    NumS:=1;S1:=tstr;
  end
  else
  begin
    NumS:=2;
    if (pos('[AND]',tstr)<>0) then ANDs:=true else ORs:=true;
    S1:=Copy(Tstr,1,(pos('[',tstr)-1));
    S2:=Copy(tstr,(pos(']',tstr)+1),length(tstr)-(pos(']',tstr)));
  end;

  If S1[1]='{' then
  begin
    S1dna:=true;
    S1:= copy(s1,2,length(s1)-2);
  End;
  If S2[1]='{' then
  begin
    S2dna:=true;
    S2:= copy(s2,2,length(s2)-2);
  end;

 {} Writeln('S1=',S1,'  S2=',S2);
 {} contin:=readkey;
  {Get input file}
  Writeln;
    {Open the main Fasta File}
  Infilename := '';
  Writeln;
  Writeln('Press <Enter> to choose a Fasta file to input');
  repeat Contin:=Readkey until contin=#13;
  GetExistingSubDirect(DirPath,True,SubDirPath);

  GetOldFile(subdirpath+'*.FA',InFilename);
  If INfilename = '' then Exit;
  Assign(Infile, subdirpath+Infilename);
  Reset(infile);
  Writeln;

  {Get option for output}
  Writeln;
  Writeln('Press <Enter> to choose a Fasta file to Output');
  repeat Contin:=Readkey until contin=#13;
  GetExistingSubDirect(DirPath,True,SubDirPath);
  GetNewFile(subdirpath+'*.FA',OutFileName,'FA');
  Assign(OutFile,subdirpath+OutFilename);
  Rewrite(OutFile);

  Writeln;
  Writeln('Analyzing the sequences.  Be Patient...');
  Write('No. of sequences analyzed: ');
  ScrX:=WhereX; Scry:=WhereY;

  repeat readln(infile, Tstr) until tstr[1]='>';
  CurHead:='';
  NextHead:=Tstr;
  TotSeq:=0;
  Repeat
    Inputseq(infile);
    S1found:=false;S2Found:=False;

    If S1dna then
    begin
      If FindMatch(1,seqlen,S1,0)<>0 then S1Found:=true
    end
    Else If pos(S1,Curhead)<>0 then S1Found:=true;

    IF Nums=2 Then
    Begin
      If S2dna then
      begin
        If FindMatch(1,seqlen,S1,0)<>0 then S2Found:=true
      end
      Else If pos(S2,Curhead)<>0 then S2Found:=true
    End;

    If Nums=1 then
    begin
      if S1found then outputseq(1,seqlen,outfile);Inc(NumOut)
    end
    Else
    Begin
      If ANDs then If (S1found and S2found) then
      begin
        outputseq(1,seqlen,outfile);Inc(NumOut)
      end
      Else If ORs then IF (S1found or s2found) then
      begin
        outputseq(1,seqlen,outfile);Inc(NumOut)
      end
    End;

    Inc(TotSeq);
    IF ((TotSeq mod 1000) = 0) then
    begin
      GotoXY(ScrX,ScrY);
      Write(TotSeq:7);
    end;
  Until EOf(Infile);
  Close(infile);close(outfile);
  Writeln;writeln;
  Writeln(totseq,' sequences were analyzed from ',infilename);
  Writeln(Numout,' sequences were output to file ',outfilename);
  Writeln;
  Continue(contin);

  Clrscr;

End;
{****************************************************************}
Procedure FastaToExcel;
Var
  I,J,K:longint;
  bC,eC,Err:Integer;
  SbC,SeC,Fmem:string[4];
  Tstr:String;
  ScrX,ScrY:Byte;
Begin
  Clrscr;
  Writeln('This routine inputs a FASTA file after V, J and CD3 analysis ');
  Writeln('and outputs an Excel file with all of the analysis data and sequences.');
  Writeln;
  Writeln('Press "Enter" to continue.');

  Infilename := '';
  Writeln;
  Writeln('Press <Enter> to choose a Fasta file to input');
  repeat Contin:=Readkey until contin=#13;
  GetExistingSubDirect(DirPath,True,SubDirPath);

  GetOldFile(subdirPath+'*.FA',InFilename);
  If INfilename = '' then Exit;
  Assign(Infile, subdirpath+Infilename);
  Reset(infile);
  Writeln;

  GetStrain;

  {Get option for output}
  Writeln;
  Outfilename:=Copy(infilename,1,(pos('.',infilename))) + 'TXT';
  Assign(OutFile,subdirpath+OutFilename);
  Rewrite(OutFile);
  {File Header}
  Write(outfile,'Seq No.,Exp Name,Seq. Name,Origin');
  Writeln(outfile,',Strain,Vfam,Vsub,J,Function,CDR3 Protein,CDR3 DNA,CDR3 Base Nos.,Sequence');
  {First Seq header}
  repeat readln(infile, Tstr) until tstr[1]='>';
  CurHead:='';
  NextHead:=Tstr;
  I:=0;
  Writeln;
  Writeln('Analyzing the sequences.  Be Patient...');
  Write('No. of sequences analyzed: ');
  ScrX:=WhereX; Scry:=WhereY;

  Repeat
    InputSeq(Infile);
    Inc(I);Write(Outfile,I,','); {Seq Num}
    {parse out the header}
    Write(Outfile,copy(CurHead,2,5),',');{Experiment Name}
    Write(Outfile,Copy(CurHead,7,9),',');{Seq Name}
    Write(Outfile,Copy(CurHead,17,8),',');{Seq Origin}

    {V and J if  identified}
    If (copy(CurHead,26,4)= 'No V') then
       Write(outfile,Vstrain,',No V,,')
    else {Write the Vfamily and family Member}
    Begin
       Write(Outfile,VStrain,',TR',Copy(CurHead,28,4),',');{Strain and Vfam}
       Fmem:= Copy(CurHead,32,4);
       If Fmem<>'    'Then
       Begin
         If FMem[4] = ' ' then FMem:= Copy(Fmem,1,3);
         If Fmem[3] = ' ' then FMem:= Copy(Fmem,1,2);
         Write(Outfile,FMem,',')
       end
       else Write(outfile,'Unk,');{Vfam Mem}
    End;

    If (copy(CurHead,37,4)= 'No J') then
       Write(outfile,'No J,')
    else Write(Outfile,'TR',Copy(CurHead,39,6),',');{J Nam }

    {Function}
    Case CurHead[46] of
    'P':Write(outfile,'1-Prod');
    'T':Write(outfile,'2-Term');
    'O':Write(outfile,'3-OOF');
    'U':Write(outfile,'4-Unk');
    End;
    Write(outfile,',');

    {Cdr3 Protein}
    Case CurHead[46] of
    'P','T':Write(outfile,Copy(CurHead,58,(Length(CurHead)-57)));
    End;
    Write(outfile,',');

    J:=48; Sbc:='';
    repeat
      Sbc:=Sbc+CurHead[J];Inc(J);
    Until (CurHead[J]=' ') or (J=52);
    Val(SbC,bC,err);

    J:=53; Sec:='';
    repeat
      Sec:=Sec+CurHead[J];Inc(J);
    Until (CurHead[J]=' ') or (J=57);
    Val(SeC,eC,err);

    {Cdr3 Dna}
    Case CurHead[46] of
    'P','T','O':for J:= bC to eC do Write(outfile,ChrA[J]);
    End;
    Write(outfile,',');

    {Cdr3 Bases Nos.}
    If (bC <> 0) then Write(outfile,bC:4,'-') else Write(outfile,' ?? -');
    IF (eC <> 0) then Write(outfile,eC:4) else Write(outfile,' ?? ');
    Write(outfile,',');

    {Output the Sequence}
     For J:= 1 to Seqlen do Write(outfile,ChrA[J]);
    Writeln(outfile);
    IF ((i mod 1000) = 0) then
    begin
      Gotoxy(ScrX,ScrY);
      Write(I:7);
    end;
  Until EOF(infile);
  Close(Infile);Close(OutFile);
  Writeln;
  Writeln;
  Writeln(I,' sequences have been analyzed from file ',infilename);
  Writeln('Your results are in : ',outfilename);
  Writeln;
  Continue(contin);
  Clrscr;
End;
{****************************************************************}
Procedure RawExcelFile;
{Convert Raw 454 file to Excel txt file}
Var
  Head454,tstr:string;
  M,L,I :integer;
  S:longint;
  EOseq:Boolean;

Begin
  Infilename := '';
  Writeln;
  Writeln('Press <Enter> to choose a Fasta file to input');
  repeat Contin:=Readkey until contin=#13;
  GetExistingSubDirect(DirPath,True,SubDirPath);

  GetOldFile(subdirPath+'*.FA',InFilename);
  If INfilename = '' then Exit;
  Assign(Infile, subdirpath+Infilename);
  Reset(infile);

  Writeln;
  Outfilename:=Copy(infilename,1,(pos('.',infilename))) + 'TXT';
  Assign(OutFile,subdirpath+OutFilename);
  Rewrite(OutFile);


  Repeat Readln(infile,Tstr) until Tstr[1]='>';
  Writeln(outfile, 'Seq No., Exp. No., Seq No., Sequence');
  S:=0;

  While NOT Eof(infile) Do
  Begin
    FillChar(ChrA,1000,#0);
    M:=1; {Counter for next open spot in ChrA}
    EOseq:=false;
    inc(S);
    Write(outfile,s,',',Copy(Tstr,2,7),',',copy(Tstr,11,5),',');
    readln(infile, tstr);

    {assemble seq convert to complement}
    Repeat
      L:=Length(Tstr);                      {Length of Current line}
      Move(tstr[1],CHRA[m],L);              {Add Current line to array}
      M:=M + L;                             {Next available base in ChrA}
      readln(infile,tstr);
    Until (Tstr[1]='>') or Eof(Infile);
    Seqlen:=M-1;
    CompFlip;

    For I:=1 to Seqlen do Write(outfile, ChrA[I]);
    Writeln(outfile);
  End;

  Close(infile);
  Close(outfile);

End;{RawExcelfile}
{****************************************************************}
Procedure GetAnalysisOption;
{Get the user option, clear memory, read in databases}
Begin
  Clrscr;
  OKtoQuit:=false;
  Writeln('Options');
  Writeln('1. Clean Up the Sequences (DO THIS FIRST!!).');
  Writeln('2. Identify V, J and CDR3 for Each Sequence (DO THIS THIRD!!).');
  Writeln('3. Extract a Subset of Sequences.');
  Writeln('4. Convert Fasta file to Excel file.');
  Writeln('5. Create a summary VJ Excel Table.');
  {Writeln('$. Make MKset File;');}
  Writeln('<ESC>. Quit.');

  Repeat option:=readkey until (option in ['1'..'5',#27,'$','X']);

  If option =#27 then OKtoQuit:=true;
  Clrscr;
End;
{****************************************************************}
{****************************************************************}
{Main Program}
Begin
  Initialize;
  GetWorkingDirectory;
  Repeat GetDataOption Until Dataset in ['1',#27];
  If Dataset = #27 then Halt;
  {database files must be in the working directory}
  LoadVDataBase;
  LoadJDataBase;
  repeat
    GetAnalysisOption;
    If not OKtoQuit then
    Begin
      Case Option of
      '1': CleanUpSeq;
      '2': IdentifyVJ;
      '3': ExtractSubset;
      '4': FastaToExcel;
      '5': MakeExcelVJTable;
      '$': MakeMKSetFile;{hidden procedure}
      'X': RawExcelFile;{hidden procedure}
      End;
    End;
  until OKtoQuit;
  Close(Infile);
End.
